from pathlib import Path
from shutil import copytree, rmtree
from argparse import ArgumentParser

def main():
    parser = ArgumentParser()
    parser.add_argument("ghidra_src", help="Path to the Ghidra source code")
    args = parser.parse_args()
    ghidra_src = Path(args.ghidra_src)
    if not ghidra_src.is_dir():
        raise FileNotFoundError(f"Not a directory: {ghidra_src}")
    processors = ghidra_src.joinpath("Ghidra", "Processors")
    if not ghidra_src.is_dir():
        raise FileNotFoundError(f"Not found: {processors}")
    target_root = Path.cwd().joinpath("python", "icicle", "Ghidra", "Processors")
    if target_root.exists():
        rmtree(target_root)
    for processor in processors.glob("*"):
        if not processor.is_dir():
            continue
        name = processor.name
        languages = processor.joinpath("data", "languages")
        if not languages.is_dir():
            print(f"Processor skipped: {name}")
            continue
        print(f"Processor: {name}")
        processor_root = target_root.joinpath(name, "data", "languages")
        copytree(languages, processor_root, dirs_exist_ok=True)

if __name__ == "__main__":
    main()


from icicle import *

def main():
    emu = Icicle("x86_64", jit=True, optimize_block=True, tracing=True)

    instructions = bytes.fromhex("41 C1 EA 07 41 83 E2 1F 74 08 44 89 D0 48 89 54 C6 08 49 83 C1 04 4C 89 0E 4C 89 C9 44 8B 11 44 89 D0 F7 D0 49 89 C9 A8 03 0F 84 88 F6 FF FF EB 4C".replace(" ", ""))

    addr = 0x140001A73
    heap = 0x71000
    emu.mem_map(heap, 0x1000, MemoryProtection.ReadWrite)
    emu.mem_write(heap + 4, 0x1337.to_bytes(4, "little"))
    emu.mem_map(addr & ~0xFFF, 0x1000, MemoryProtection.ExecuteRead)
    emu.mem_write(addr, instructions)
    emu.reg_write("r9", heap)
    emu.reg_write("r10", 0x13)
    emu.reg_write("rip", addr)
    emu.reg_write("rsi", heap + 0x100)

    for i in range(11):
        rip = emu.reg_read("rip")
        rcx = emu.reg_read("rcx")
        r9 = emu.reg_read("r9")
        print(f"[{i}] RIP: {hex(rip)}, RCX: {hex(rcx)}, R9: {hex(r9)}")

        if rip == 0x140001A8F:
            emu.reg_write("r9", 0x13370900)

        emu.step(1)

        if rip == 0x140001A9A:
            assert rcx == r9, f"expected rcx({hex(rcx)}) == r9({hex(r9)})"

    print("Everything works!")

if __name__ == "__main__":
    main()


from typing import List, Dict, Tuple
from enum import Enum

class MemoryProtection(Enum):
    NoAccess = ...
    ReadOnly = ...
    ReadWrite = ...
    ExecuteOnly = ...
    ExecuteRead = ...
    ExecuteReadWrite = ...

class MemoryExceptionCode(Enum):
    Unallocated = ...
    Unmapped = ...
    UnmappedRegisters = ...
    Uninitialized = ...
    ReadViolation = ...
    WriteViolation = ...
    ExecViolation = ...
    ReadWatch = ...
    WriteWatch = ...
    Unaligned = ...
    OutOfMemory = ...
    SelfModifyingCode = ...
    AddressOverflow = ...
    Unknown = ...

class RunStatus(Enum):
    Running = ...
    InstructionLimit = ...
    Breakpoint = ...
    Interrupted = ...
    Halt = ...
    Killed = ...
    Deadlock = ...
    OutOfMemory = ...
    Unimplemented = ...
    UnhandledException = ...

class ExceptionCode(Enum):
    NoException = ...
    InstructionLimit = ...
    Halt = ...
    Sleep = ...
    Syscall = ...
    CpuStateChanged = ...
    DivisionException = ...
    ReadUnmapped = ...
    ReadPerm = ...
    ReadUnaligned = ...
    ReadWatch = ...
    ReadUninitialized = ...
    WriteUnmapped = ...
    WritePerm = ...
    WriteWatch = ...
    WriteUnaligned = ...
    ExecViolation = ...
    SelfModifyingCode = ...
    OutOfMemory = ...
    AddressOverflow = ...
    InvalidInstruction = ...
    UnknownInterrupt = ...
    UnknownCpuID = ...
    InvalidOpSize = ...
    InvalidFloatSize = ...
    CodeNotTranslated = ...
    ShadowStackOverflow = ...
    ShadowStackInvalid = ...
    InvalidTarget = ...
    UnimplementedOp = ...
    ExternalAddr = ...
    Environment = ...
    JitError = ...
    InternalError = ...
    UnmappedRegister = ...
    UnknownError = ...

class Icicle:
    def __init__(self, architecture: str, *,
                 jit = True,
                 jit_mem = True,
                 shadow_stack = False,
                 recompilation = True,
                 track_uninitialized = False,
                 optimize_instructions = True,
                 optimize_block = False,
                 tracing = False,
                 ) -> None: ...

    @property
    def architecture(self) -> str: ...

    @property
    def exception_code(self) -> ExceptionCode: ...

    @property
    def exception_value(self) -> int: ...

    icount: int

    icount_limit: int

    pc: int

    sp: int

    # TODO: API to get memory information?

    def mem_map(self, address: int, size: int, protection: MemoryProtection): ...

    def mem_unmap(self, address: int, size: int): ...

    def mem_protect(self, address: int, size: int, protection: MemoryProtection): ...

    def mem_read(self, address: int, size: int) -> bytes: ...

    def mem_write(self, address: int, data: bytes) -> None: ...

    def reg_list(self) -> Dict[str, Tuple[int, int]]: ...

    def reg_offset(self, name: str) -> int: ...

    def reg_size(self, name: str) -> int: ...

    def reg_read(self, name: str) -> int: ...

    def reg_write(self, name: str, value: int) -> None: ...

    def reset(self): ...

    def run(self) -> RunStatus: ...

    def run_until(self, address: int) -> RunStatus: ...

    def step(self, count: int) -> RunStatus: ...

    def add_breakpoint(self, address: int) -> bool: ...

    def remove_breakpoint(self, address: int) -> bool: ...

def architectures() -> List[str]: ...

class MemoryException(Exception):
    def __init__(self, message: str, code: MemoryExceptionCode):
        super().__init__(message)
        self.code = code

    def __str__(self):
        return f"{super().__str__()}: {self.code}"

def __ghidra_init():
    import os
    for path in __path__ + [os.getenv("GHIDRA_SRC", ".")]:
        processors_dir = os.path.join(path, "Ghidra", "Processors")
        if os.path.isdir(processors_dir):
            os.putenv("GHIDRA_SRC", path)
            os.environ["GHIDRA_SRC"] = path
            return
    raise FileNotFoundError("Ghidra processor definitions not found")

__ghidra_init()

# NOTE: This overrides the stubs at runtime with the actual implementation
from .icicle import *


"""
Scenarios:
- Non-executable memory being executed
  - What should happen when the memory doesn't exist at all?
- Invalid instruction at the start of the block
- Invalid instruction in the middle of a block
- IP alignment error (non-x86)
"""

from icicle import *

def nx():
    vm = Icicle("x86_64", jit=False)
    page = 0x10000
    vm.mem_map(page, 0x1000, MemoryProtection.ReadOnly)
    vm.mem_write(page, b"\x90\xC3")
    vm.reg_write("rip", page)
    status = vm.run()
    print(status)
    print(vm.exception_code, hex(vm.exception_value))
    # We expect an execution violation at rip
    assert vm.exception_code == ExceptionCode.ExecViolation
    assert vm.exception_value == page

def inv_start():
    vm = Icicle("x86_64", jit=False)
    page = 0x10000
    vm.mem_map(page, 0x1000, MemoryProtection.ExecuteRead)
    vm.mem_write(page, b"\xFF\xFF\x90")
    vm.reg_write("rip", page)
    status = vm.run()
    print(status)
    print(vm.exception_code, hex(vm.exception_value))
    # We expect an invalid instruction at rip
    assert vm.exception_code == ExceptionCode.InvalidInstruction
    assert vm.exception_value == page

def inv_middle():
    vm = Icicle("x86_64", jit=False)
    page = 0x10000
    vm.mem_map(page, 0x1000, MemoryProtection.ExecuteRead)
    vm.mem_write(page, b"\x90\x90\xFF\xFF\x90")
    vm.reg_write("rip", page)
    status = vm.run()
    print(status)
    print(vm.exception_code, hex(vm.exception_value))
    # We expect an invalid instruction at rip+2
    assert vm.exception_code == ExceptionCode.InvalidInstruction
    assert vm.exception_value == page + 2

def main():
    print("=== NX ===")
    nx()
    print("=== Invalid instruction (block start) ===")
    inv_start()
    print("=== Invalid instruction (block middle) ===")
    inv_middle()
    print("\nSUCCESS")

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
from os import getenv
from shutil import which
from setuptools import find_packages, setup
from setuptools_rust import RustExtension

if __name__ == "__main__":
    if "sdist" not in sys.argv and which("cargo") is None:
        raise FileNotFoundError(f"Rust not found, visit https://rustup.rs for installation instructions")

    ref_name = getenv("GITHUB_REF_NAME")
    if getenv("GITHUB_REF", "").startswith("refs/tags/") and ref_name:
        from pkg_resources import parse_version
        try:
            parse_version(ref_name)
            print(f"injecting version = {ref_name} into setup.cfg")
            with open("setup.cfg", "r") as f:
                lines = f.readlines()
            with open("setup.cfg", "w") as f:
                for line in lines:
                    if line.startswith("version = "):
                        line = f"version = {ref_name}\n"
                    f.write(line)
        except Exception:
            pass

    setup(
        rust_extensions=[RustExtension("icicle.icicle")],
    )



import icicle
import keystone
import capstone

def assemble(code: str, addr: int = 0) -> bytes:
    ks = keystone.Ks(keystone.KS_ARCH_X86, keystone.KS_MODE_64)
    encoding, count = ks.asm(code, addr)
    if encoding is None:
        raise keystone.KsError("no encoding")
    return bytes(encoding)

def disassemble(code: bytes, addr: int = 0, max_count = 1000) -> str:
    result = ""
    cs = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64)
    count = 0
    while count < max_count:
        count += 1
        # TODO: fix on invalid instruction
        for address, size, mnemonic, op_str in cs.disasm_lite(code, addr, 1):
            if len(result) > 0:
                result += "\n"
            result += hex(address) + "|"
            result += mnemonic
            if op_str:
                result += " "
                result += op_str
            code = code[size:]
            if len(code) == 0:
                return result
            addr += size
    return result

def old_test():
    buf = assemble("mov eax, ebx\nnop")
    print(disassemble(buf, 0))
    print(assemble("mov eax, ebx").hex())
    print("Hello world!")
    print(dir(icicle))
    print(icicle.MemoryProtection.ReadOnly)
    print(f"Architectures: {icicle.architectures()}")
    vm = icicle.Icicle("x86_64")
    print(vm)

    for name, (offset, size) in vm.reg_list().items():
        value = vm.reg_read(name.lower())
        print(f"{name}[{size}] @ {offset} = {value}")

    addr = 0x10000
    vm.mem_map(addr, 0x1000, icicle.MemoryProtection.NoAccess)
    try:
        vm.mem_protect(addr, 0x2000, icicle.MemoryProtection.ExecuteRead)
    except icicle.MemoryException as x:
        print("MemoryException working!")
        message = x.args[0]
        print(message, x.code)
    except Exception as x:
        print(x, type(x), isinstance(x, icicle.MemoryException))
    vm.mem_protect(addr, 0x1000, icicle.MemoryProtection.ExecuteRead)
    vm.mem_write(addr, b"\x90\x90\x90")
    data = vm.mem_read(addr, 4)
    print(data, type(data))

def main():
    old_test()

    print("")
    vm = icicle.Icicle("x86_64", jit=False)
    print(vm)

    addr = 0x10000
    #vm.mem_map(0, 0x1000, icicle.MemoryProtection.ReadWrite)
    vm.mem_map(addr, 0x1000, icicle.MemoryProtection.ExecuteRead)
    code = assemble("""
mov eax, 42
mov qword ptr [rax], 0x5
nop
""", addr)
    print("code:")
    print(disassemble(code, addr))
    vm.mem_write(addr, code)
    data = vm.mem_read(addr, len(code))
    assert data == code

    vm.reg_write("rip", addr)
    status = vm.run_until(addr + len(code))
    print(status)
    if status == icicle.RunStatus.UnhandledException:
        print(f"code: {vm.exception_code}, value: {hex(vm.exception_value)}")
        print(f"rip: {hex(vm.reg_read('rip'))} (OK)")
    else:
        print(f"rax: {hex(vm.reg_read('rax'))}")
        assert False, "Should be unreachable"

if __name__ == "__main__":
    main()
